@@ -31,9 +31,12 @@ use derive_more::{Display, From};
 use encoding_rs::mem;
 use im::HashMap as ImHashMap;
 use im::OrdMap;
+use lazy_static::lazy_static;
 use murmurhash3::murmurhash3_x64_128;
 use ordered_float::OrderedFloat;
+use regex::Regex;
 use serde::{Deserialize, Serialize};
+use snafu::Snafu;
 use std::cmp::{min, Reverse};
 use std::collections::{BTreeMap, HashMap};
 use std::convert::From;
@@ -43,13 +46,35 @@ use std::fmt::{Display, Formatter};
 use std::net::{SocketAddr, ToSocketAddrs};
 use std::ops::Index;
 use std::vec;
+use tokio_rustls::rustls;
+use tokio_rustls::rustls::{RootCertStore, ServerCertVerified, ServerCertVerifier, TLSError};
+use tokio_rustls::webpki::DNSNameRef;
 use uuid::Uuid;
+
 #[macro_use]
 extern crate shrinkwraprs;
 
 #[macro_use]
 extern crate derive_new;
 
+#[macro_use]
+extern crate num_derive;
+
+type PravegaNodeUriParseResult<T> = std::result::Result<T, PravegaNodeUriParseError>;
+
+#[derive(Debug, Snafu)]
+pub enum PravegaNodeUriParseError {
+    #[snafu(display("Could not parse uri due to {}", error_msg))]
+    ParseError { error_msg: String },
+}
+
+#[derive(Debug, PartialEq, Default)]
+struct PravegaNodeUriParts {
+    scheme: Option<String>,
+    domain_name: Option<String>,
+    port: Option<u16>,
+}
+
 #[derive(From, Shrinkwrap, Debug, Clone, Hash, PartialEq, Eq)]
 pub struct PravegaNodeUri(pub String);
 
@@ -75,22 +100,95 @@ impl From<SocketAddr> for PravegaNodeUri {
 impl PravegaNodeUri {
     pub fn to_socket_addr(&self) -> SocketAddr {
         // to_socket_addrs will resolve hostname to ip address
-        let mut addrs_vec: Vec<_> = self
-            .0
-            .to_socket_addrs()
-            .expect("Unable to resolve domain")
-            .collect();
-        addrs_vec.pop().expect("get the first SocketAddr")
+        match PravegaNodeUri::uri_parts_from_string(self.to_string()) {
+            Ok(uri_parts) => {
+                let mut addrs_vec: Vec<_> =
+                    format!("{}:{}", uri_parts.domain_name.unwrap(), uri_parts.port.unwrap())
+                        .to_socket_addrs()
+                        .expect("Unable to resolve domain")
+                        .collect();
+                addrs_vec.pop().expect("get the first SocketAddr")
+            }
+            Err(e) => panic!("{}", e),
+        }
     }
 
     pub fn domain_name(&self) -> String {
-        let parts: Vec<_> = self.0.split(':').collect();
-        parts[0].to_string()
+        match PravegaNodeUri::uri_parts_from_string(self.to_string()) {
+            Ok(uri_parts) => uri_parts.domain_name.expect("uri missing domain name"),
+            Err(e) => panic!("{}", e),
+        }
     }
 
     pub fn port(&self) -> u16 {
-        let parts: Vec<_> = self.0.split(':').collect();
-        parts[1].parse::<u16>().expect("parse port to u16")
+        match PravegaNodeUri::uri_parts_from_string(self.to_string()) {
+            Ok(uri_parts) => uri_parts.port.expect("parse port to u16"),
+            Err(e) => panic!("{}", e),
+        }
+    }
+
+    /// Return Result of the uri scheme or empty string if no scheme was specified
+    pub fn scheme(&self) -> PravegaNodeUriParseResult<String> {
+        match PravegaNodeUri::uri_parts_from_string(self.to_string()) {
+            Ok(sa) => match sa.scheme {
+                Some(scheme) => Ok(scheme),
+                _ => Ok("".to_string()),
+            },
+            Err(e) => Err(e),
+        }
+    }
+
+    /// verifies the uri is well-formed (contains at least host:port, with optional scheme:// prefix)
+    ///
+    pub fn is_well_formed(uri: String) -> bool {
+        match PravegaNodeUri::uri_parts_from_string(uri) {
+            Ok(uri_parts) => uri_parts.port.is_some() && uri_parts.domain_name.is_some(),
+            Err(_) => false,
+        }
+    }
+
+    fn uri_parts_from_string(uri: String) -> PravegaNodeUriParseResult<PravegaNodeUriParts> {
+        lazy_static! {
+            static ref URI_RE: Regex = Regex::new(
+                r"(?x)
+            (?:(?P<scheme>[[:alnum:]]+)://)?
+            (?P<domain_name>([0-9A-Za-z\-\.]+|\[[0-9A-F\.:]+\]))
+            :
+            (?P<port>[[:digit:]]+)"
+            )
+            .unwrap();
+        }
+        let mut uri_parts: PravegaNodeUriParts = PravegaNodeUriParts::default();
+
+        // The Java client supports multiple comma separated endpoints in a single string
+        // where the first endpoint has the scheme to be applied to all endpoints.
+        // To be semi-compatible with the Java code this method accepts a string containing comma
+        // separated (or any separator) endpoints, but it uses only the first endpoint in the string.
+        // The domain name portion can be an ip name, ipv4 address or an ipv6 literal
+        // ipv6 addresses retain [] wrapper because both to_socket_addrs() and
+        // http URI from_str() require IP literals
+        let first_endpoint = match URI_RE.captures_iter(&uri).next() {
+            Some(endpoint) => endpoint,
+            _ => {
+                return Err(PravegaNodeUriParseError::ParseError {
+                    error_msg: format!("malformed uri {}", uri),
+                })
+            }
+        };
+
+        uri_parts.domain_name = Some(first_endpoint.name("domain_name").unwrap().as_str().to_string());
+        uri_parts.port = Some(
+            first_endpoint
+                .name("port")
+                .unwrap()
+                .as_str()
+                .parse::<u16>()
+                .expect("port not a valid u16"),
+        );
+        uri_parts.scheme = first_endpoint
+            .name("scheme")
+            .map(|scheme| scheme.as_str().to_string());
+        Ok(uri_parts)
     }
 }
 
@@ -128,7 +226,7 @@ pub struct Reader {
     pub name: String,
 }
 
-#[derive(Debug, Clone, Hash, PartialEq, Eq, Serialize, Deserialize)]
+#[derive(Clone, Hash, PartialEq, Eq, Serialize, Deserialize)]
 pub struct Segment {
     pub number: i64,
     pub tx_id: Option<TxId>,
@@ -158,6 +256,23 @@ impl Segment {
     pub fn is_transaction_segment(&self) -> bool {
         self.tx_id.is_some()
     }
+
+    pub fn get_epoch(&self) -> i32 {
+        (self.number >> 32) as i32
+    }
+
+    pub fn get_segment_number(&self) -> i32 {
+        self.number as i32
+    }
+}
+
+impl fmt::Debug for Segment {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("Segment")
+            .field("segment", &self.get_segment_number())
+            .field("epoch", &self.get_epoch())
+            .finish()
+    }
 }
 
 #[derive(new, Debug, Clone, Hash, PartialEq, Eq, Serialize, Deserialize)]
@@ -234,10 +349,11 @@ impl From<&str> for ScopedSegment {
             ScopedSegment::from(original_segment_name)
         } else {
             let mut tokens = NameUtils::extract_segment_tokens(qualified_name.to_owned());
-            if tokens.len() == 2 {
+            let segment_id = tokens.pop().expect("get segment id from tokens");
+            let stream_name = tokens.pop().expect("get stream name from tokens");
+
+            if tokens.is_empty() {
                 // scope not present
-                let segment_id = tokens.pop().expect("get segment id from tokens");
-                let stream_name = tokens.pop().expect("get stream name from tokens");
                 ScopedSegment {
                     scope: Scope {
                         name: String::from(""),
@@ -249,8 +365,6 @@ impl From<&str> for ScopedSegment {
                     },
                 }
             } else {
-                let segment_id = tokens.pop().expect("get segment id from tokens");
-                let stream_name = tokens.pop().expect("get stream name from tokens");
                 let scope = tokens.pop().expect("get scope from tokens");
                 ScopedSegment {
                     scope: Scope { name: scope },
@@ -327,7 +441,7 @@ impl Display for WriterId {
     }
 }
 
-#[derive(Debug, Clone, Hash, PartialEq, Eq)]
+#[derive(Debug, Clone, Hash, PartialEq, Eq, FromPrimitive)]
 pub enum ScaleType {
     FixedNumSegments = 0,
     ByRateInKbytesPerSec = 1,
@@ -348,12 +462,12 @@ impl Default for Scaling {
             scale_type: ScaleType::FixedNumSegments,
             min_num_segments: 1,
             scale_factor: 1,
-            target_rate: 1000,
+            target_rate: 0,
         }
     }
 }
 
-#[derive(Debug, Clone, Hash, PartialEq, Eq)]
+#[derive(Debug, Clone, Hash, PartialEq, Eq, FromPrimitive)]
 pub enum RetentionType {
     None = 0,
     Time = 1,
@@ -386,7 +500,7 @@ impl Default for Retention {
     fn default() -> Self {
         Retention {
             retention_type: RetentionType::None,
-            retention_param: std::i64::MAX,
+            retention_param: 0,
         }
     }
 }
@@ -396,6 +510,7 @@ pub struct StreamConfiguration {
     pub scoped_stream: ScopedStream,
     pub scaling: Scaling,
     pub retention: Retention,
+    pub tags: Option<Vec<String>>,
 }
 
 #[derive(new, Debug, Clone)]
@@ -674,6 +789,21 @@ pub struct SegmentInfo {
     pub last_modified_time: i64,
 }
 
+/// This struct is used to to skip cert verifications.
+pub struct NoVerifier;
+
+impl ServerCertVerifier for NoVerifier {
+    fn verify_server_cert(
+        &self,
+        _roots: &RootCertStore,
+        _presented_certs: &[rustls::Certificate],
+        _dns_name: DNSNameRef,
+        _ocsp_response: &[u8],
+    ) -> Result<ServerCertVerified, TLSError> {
+        Ok(ServerCertVerified::assertion())
+    }
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
@@ -731,6 +861,68 @@ mod test {
         assert_eq!(uri.domain_name(), "localhost".to_string());
         assert_eq!(uri.port(), 9090);
         assert_eq!(uri.to_socket_addr(), socket_addr);
+
+        assert_eq!(
+            PravegaNodeUri::uri_parts_from_string("127.0.0.1:9090".to_string()).unwrap(),
+            PravegaNodeUriParts {
+                scheme: None,
+                domain_name: Some("127.0.0.1".to_string()),
+                port: Some(9090)
+            }
+        );
+
+        let uri_with_scheme = PravegaNodeUri("tls://127.0.0.1:9090".to_string());
+        assert_eq!(
+            PravegaNodeUri::uri_parts_from_string(uri_with_scheme.to_string()).unwrap(),
+            PravegaNodeUriParts {
+                scheme: Some("tls".to_string()),
+                domain_name: Some("127.0.0.1".to_string()),
+                port: Some(9090),
+            }
+        );
+        assert!(PravegaNodeUri::is_well_formed(uri_with_scheme.to_string()));
+        assert_eq!(uri_with_scheme.port(), 9090);
+
+        // test a multi-endpoint uri that java client claims to support in client/src/main/java/io/pravega/client/ClientConfig.java
+        let uri_with_scheme =
+            PravegaNodeUri("ssl://127.0.0.1:9090,127.0.0.1:9091,127.0.0.1:9092".to_string());
+        assert_eq!(
+            PravegaNodeUri::uri_parts_from_string(uri_with_scheme.to_string()).unwrap(),
+            PravegaNodeUriParts {
+                scheme: Some("ssl".to_string()),
+                domain_name: Some("127.0.0.1".to_string()),
+                port: Some(9090),
+            }
+        );
+
+        assert!(PravegaNodeUri::uri_parts_from_string("tls://127.0.0.1://9090".into()).is_err());
+        assert!(!PravegaNodeUri::is_well_formed("tls://127.0.0.1://9090".into()));
+        assert!(PravegaNodeUri("tls://127.0.0.1://9090".to_string())
+            .scheme()
+            .is_err());
+        assert_eq!(
+            PravegaNodeUri("tls://127.0.0.1:9090".to_string())
+                .scheme()
+                .unwrap(),
+            "tls".to_string()
+        );
+        assert!(PravegaNodeUri("".to_string()).scheme().is_err());
+
+        // test ipv6 literal
+        assert!(
+            PravegaNodeUri("tcps://[1762:0:0:0:0:B03:1:AF18]:12345".to_string())
+                .scheme()
+                .is_ok()
+        );
+
+        assert_eq!(
+            PravegaNodeUri::uri_parts_from_string("tcps://[::1]:12345".to_string()).unwrap(),
+            PravegaNodeUriParts {
+                scheme: Some("tcps".to_string()),
+                domain_name: Some("[::1]".to_string()),
+                port: Some(12345),
+            }
+        );
     }
 
     #[test]
